/* See the end of this file for copyright and license terms. */

#include <catch2/catch.hpp>
#include <errno.h>

#include <neo.h>

extern "C" struct nstr2i_positive_case {
	const char *s;
	i64 i;
	int radix;
};

TEST_CASE( "nstr2i: Parse numbers", "[string/nstr2x.c]" )
{
	struct nstr2i_positive_case positives[] = {
		{"0",			0,				2	},
		{"0",			0,				36	},
		{"1y2p0ij32e8e7",	0x7fffffffffffffff,		36	},
		{"-1y2p0ij32e8e8",	(i64)0x8000000000000000,	36	},
		{"1Y2P0IJ32E8E7",	0x7fffffffffffffff,		36	},
		{"-1Y2P0IJ32E8E8",	(i64)0x8000000000000000,	36	},
		{"0x7fffffffffffffff",	0x7fffffffffffffff,		0	},
		{"-0x8000000000000000",	(i64)0x8000000000000000,	0	},
		{"69420",		69420,				0	},
		{"+69420",		69420,				0	},
		{"-69420",		-69420,				0	},
		{"0xff",		0xff,				0	},
		{"0XFF",		0xff,				0	},
		{"+0xff",		0xff,				0	},
		{"+0XFF",		0xff,				0	},
		{"-0xff",		-0xff,				0	},
		{"-0XFF",		-0xff,				0	},
		{"0b11111111",		0xff,				0	},
		{"0B11111111",		0xff,				0	},
		{"0o77777777",		077777777,			0	},
		{"0O77777777",		077777777,			0	},
		{"069420",		69420,				0	}
	};

	struct nstr2i_positive_case *current_case = &positives[0];
	do {
		error err;
		nstr_t *s = nstr(current_case->s, nil);
		i64 i = nstr2i(s, current_case->radix, &err);

		REQUIRE( i == current_case->i );
		REQUIRE( errnum(&err) == 0 );
	} while (++current_case < &positives[0] + sizeof(positives) / sizeof(positives[0]));
}

TEST_CASE( "nstr2i: Error if number too low", "[string/nstr2x.c]" )
{
	error err;
	nstr_t *s = nstr("-8000000000000001", nil);
	i64 i = nstr2i(s, 16, &err);

	nstr_t *expected_msg = nstr("Number out of range", nil);
	nstr_t *actual_msg = errmsg(&err);

	REQUIRE( i == -0x8000000000000000 );
	REQUIRE( errnum(&err) == ERANGE );
	REQUIRE( nstreq(expected_msg, actual_msg, nil) );

	nput(expected_msg);
	nput(actual_msg);
	nput(s);
}

TEST_CASE( "nstr2i: Error if number too high", "[string/nstr2x.c]" )
{
	error err;
	nstr_t *s = nstr("8000000000000000", nil);
	i64 i = nstr2i(s, 16, &err);

	nstr_t *expected_msg = nstr("Number out of range", nil);
	nstr_t *actual_msg = errmsg(&err);

	REQUIRE( i == 0x7fffffffffffffff );
	REQUIRE( errnum(&err) == ERANGE );
	REQUIRE( nstreq(expected_msg, actual_msg, nil) );

	nput(expected_msg);
	nput(actual_msg);
	nput(s);
}

TEST_CASE( "nstr2i: Error if base too low", "[string/nstr2x.c]" )
{
	error err;
	nstr_t *s = nstr("420", nil);
	i64 i = nstr2i(s, -1, &err);

	nstr_t *expected_msg = nstr("Numerical base out of range", nil);
	nstr_t *actual_msg = errmsg(&err);

	REQUIRE( i == 0 );
	REQUIRE( nstreq(expected_msg, actual_msg, nil) );
	REQUIRE( errnum(&err) == EINVAL );

	nput(expected_msg);
	nput(actual_msg);
	nput(s);
	errput(&err);
}

TEST_CASE( "nstr2i: Error if base is 1", "[string/nstr2x.c]" )
{
	error err;
	nstr_t *s = nstr("420", nil);
	i64 i = nstr2i(s, 1, &err);

	nstr_t *expected_msg = nstr("Numerical base out of range", nil);
	nstr_t *actual_msg = errmsg(&err);

	REQUIRE( i == 0 );
	REQUIRE( nstreq(expected_msg, actual_msg, nil) );
	REQUIRE( errnum(&err) == EINVAL );

	nput(expected_msg);
	nput(actual_msg);
	nput(s);
	errput(&err);
}

TEST_CASE( "nstr2i: Error if base too high", "[string/nstr2x.c]" )
{
	error err;
	nstr_t *s = nstr("420", nil);
	i64 i = nstr2i(s, 37, &err);

	nstr_t *expected_msg = nstr("Numerical base out of range", nil);
	nstr_t *actual_msg = errmsg(&err);

	REQUIRE( i == 0 );
	REQUIRE( nstreq(expected_msg, actual_msg, nil) );
	REQUIRE( errnum(&err) == EINVAL );

	nput(expected_msg);
	nput(actual_msg);
	nput(s);
	errput(&err);
}

TEST_CASE( "nstr2i: Error if character not part of numerical system", "[string/nstr2x.c]" )
{
	error err;
	nstr_t *s = nstr("012", nil);
	i64 i = nstr2i(s, 2, &err);

	nstr_t *expected_msg = nstr("'2' is not a base 2 digit", nil);
	nstr_t *actual_msg = errmsg(&err);

	REQUIRE( i == 0 );
	REQUIRE( nstreq(expected_msg, actual_msg, nil) );
	REQUIRE( errnum(&err) == EINVAL );

	nput(expected_msg);
	nput(actual_msg);
	nput(s);
	errput(&err);
}

TEST_CASE( "nstr2i: Error if non-digit character", "[string/nstr2x.c]" )
{
	error err;
	nstr_t *s = nstr("0,12", nil);
	i64 i = nstr2i(s, 2, &err);

	nstr_t *expected_msg = nstr("',' is not a base 2 digit", nil);
	nstr_t *actual_msg = errmsg(&err);

	REQUIRE( i == 0 );
	REQUIRE( nstreq(expected_msg, actual_msg, nil) );
	REQUIRE( errnum(&err) == EINVAL );

	nput(expected_msg);
	nput(actual_msg);
	nput(s);
	errput(&err);
}

TEST_CASE( "nstr2i: Error if string is nil", "[string/nstr2x.c]" )
{
	error err;
	i64 i = nstr2i(nil, 2, &err);

	nstr_t *expected_msg = nstr("String is nil", nil);
	nstr_t *actual_msg = errmsg(&err);

	REQUIRE( i == 0 );
	REQUIRE( nstreq(expected_msg, actual_msg, nil) );
	REQUIRE( errnum(&err) == EFAULT );

	nput(expected_msg);
	errput(&err);
}

TEST_CASE( "nstr2i_put: Decrement refcount", "[string/nstr2x.c]" )
{
	error err;
	nstr_t *s = nstr("420", nil);
	nget(s);
	i64 i = nstr2i_put(s, 10, &err);

	REQUIRE( i == 420 );
	REQUIRE( nref_count(s) == 1 );

	nput(s);
}

TEST_CASE( "nstr2i_put: Error if string is nil", "[string/nstr2x.c]" )
{
	error err;
	i64 i = nstr2i_put(nil, 2, &err);

	nstr_t *expected_msg = nstr("String is nil", nil);
	nstr_t *actual_msg = errmsg(&err);

	REQUIRE( i == 0 );
	REQUIRE( nstreq(expected_msg, actual_msg, nil) );
	REQUIRE( errnum(&err) == EFAULT );

	nput(expected_msg);
	errput(&err);
}

/*
 * This file is part of libneo.
 * Copyright (c) 2021 Fefie <owo@fef.moe>.
 *
 * libneo is non-violent software: you may only use, redistribute,
 * and/or modify it under the terms of the CNPLv6+ as found in
 * the LICENSE file in the source code root directory or at
 * <https://git.pixie.town/thufie/CNPL>.
 *
 * libneo comes with ABSOLUTELY NO WARRANTY, to the extent
 * permitted by applicable law.  See the CNPLv6+ for details.
 */
